-------------------------------------------------------------------------------
-- Concurrent Haskell Debugger 
--   ConcurrentQSemN Module
--     by Thomas Boettcher <thomas.boettcher@gmx.de>
-------------------------------------------------------------------------------

module CHD.ConcurrentQSemN
  (
    QSemN(..),
    newQSemN,		-- :: Int -> IO QSemN
    newQSemNLine,	-- :: CodePosition -> Int -> IO QSemN
    signalQSemN,	-- :: QSemN -> Int -> IO ()
    signalQSemNLine,	-- :: CodePosition -> QSemN -> Int -> IO ()
    waitQSemN,		-- :: QSemN -> Int -> IO ()
    waitQSemNLine	-- :: CodePosition -> QSemN -> Int -> IO ()
  )
  where


-------------------------------------------------------------------------------
-- IMPORTS
-------------------------------------------------------------------------------

import qualified Control.Concurrent as C
import System.Mem.Weak

import CHD.DebugInterface
import CHD.DebugMsgChan
import CHD.BaseTypes


-------------------------------------------------------------------------------
-- TYPES
-------------------------------------------------------------------------------

data QSemN = QSemN QSemNNo 
		   C.QSemN      
instance Show QSemN where
  showsPrec p (QSemN number _) = shows number
--  deriving Eq


-------------------------------------------------------------------------------
-- FUNCTIONS
-------------------------------------------------------------------------------

newQSemN :: Int -> IO QSemN
newQSemN = newQSemNLine NoPosition


newQSemNLine :: CodePosition -> Int -> IO QSemN
newQSemNLine pos x = do
  returnNewNoMVar <- C.newEmptyMVar
  debugStop <- sendDebugMsg (QSemNNewSuspend (CHD returnNewNoMVar)
					     (Quantity (toInteger x)) pos)
  C.takeMVar debugStop
  qsemnNo <- C.readMVar returnNewNoMVar
  qsemn <- C.newQSemN x
  debugStop2 <- sendDebugMsg (QSemNNew qsemnNo pos)
  C.takeMVar debugStop2
  addFinalizer qsemn (do {sendDebugMsg (QSemNDied qsemnNo) ; return ()})
  return (QSemN qsemnNo qsemn)


waitQSemN :: QSemN -> Int -> IO ()
waitQSemN = waitQSemNLine NoPosition


waitQSemNLine :: CodePosition -> QSemN -> Int -> IO ()
waitQSemNLine pos (QSemN qsemnNo qsemn) x = do
  debugStop1 <- sendDebugMsg (QSemNWaitSuspend qsemnNo pos) 
  C.takeMVar debugStop1
  C.waitQSemN qsemn x
  debugStop2 <- sendDebugMsg (QSemNWait (Quantity (toInteger x)) qsemnNo pos)
  C.takeMVar debugStop2


signalQSemN :: QSemN -> Int -> IO ()
signalQSemN = signalQSemNLine NoPosition


signalQSemNLine :: CodePosition -> QSemN -> Int -> IO ()
signalQSemNLine pos (QSemN qsemnNo qsemn) x = do
  debugStop1 <- sendDebugMsg (QSemNSignalSuspend qsemnNo pos)
  C.takeMVar debugStop1
  C.signalQSemN qsemn x
  debugStop2 <- sendDebugMsg (QSemNSignal (Quantity (toInteger x)) qsemnNo pos)
  C.takeMVar debugStop2
