-------------------------------------------------------------------------------
-- Concurrent Haskell Debugger 
--   Label Module
--     by Thomas Boettcher <thomas.boettcher@gmx.de>
-------------------------------------------------------------------------------

-- here the extra-features of the CHD-Bib are implemented.

module CHD.Label
  ( 
    chdBreakThread,		-- :: String -> IO ()
    labelProgram,		-- :: String -> IO ()
    labelThread,		-- :: String -> IO String
    labelMVar,			-- :: MVar a -> String -> IO String
    labelChan,			-- :: Chan a -> String -> IO String
    labelQSem,			-- :: QSem -> String -> IO String
    labelQSemN,			-- :: QSemN -> String -> IO String
    labelSampleVar,		-- :: SampleVar -> String -> IO String
    forkIOLabel,		-- :: String -> IO () -> IO C.ThreadId
    forkIOLabelLine		-- :: CodePosition -> String -> IO () 
				--		   -> IO C.ThreadId
  )
  where


-------------------------------------------------------------------------------
-- IMPORTS
-------------------------------------------------------------------------------

import qualified Control.Concurrent as C

import CHD.DebugInterface
import CHD.ConcurrentMain
import CHD.ConcurrentMVar
import CHD.ConcurrentChan
import CHD.ConcurrentQSem
import CHD.ConcurrentQSemN
import CHD.ConcurrentSampleVar
import CHD.DebugMsgChan
import CHD.BaseTypes

-------------------------------------------------------------------------------
-- TYPES
-------------------------------------------------------------------------------


-------------------------------------------------------------------------------
-- FUNCTIONS
-------------------------------------------------------------------------------

-- if a name is changed the oldname will be retrieved and returned.

chdBreakThread :: String -> IO ()
chdBreakThread breakCause = do
  debugStop <- sendDebugMsg (ThreadBreak breakCause)
  C.takeMVar debugStop
  return ()


labelProgram :: String -> IO ()
labelProgram name = do
  debugStop <- sendDebugMsg (ProgramLabel name)
  C.takeMVar debugStop
  return ()


labelMVar :: MVar a -> String -> IO String
labelMVar (MVar mvarNo mvar) name = do
  oldName <- C.newEmptyMVar
  debugStop <- sendDebugMsg (MVarLabel mvarNo name oldName)
  C.takeMVar debugStop
  C.takeMVar oldName


labelChan :: Chan a -> String -> IO String
labelChan (Chan chanNo channel) name = do
  oldName <- C.newEmptyMVar
  debugStop <- sendDebugMsg (ChanLabel chanNo name oldName)
  C.takeMVar debugStop
  C.takeMVar oldName


labelQSem :: QSem -> String -> IO String
labelQSem (QSem qsemNo qsem) name = do
  oldName <- C.newEmptyMVar
  debugStop <- sendDebugMsg (QSemLabel qsemNo name oldName)
  C.takeMVar debugStop
  C.takeMVar oldName


labelQSemN :: QSemN -> String -> IO String
labelQSemN (QSemN qsemnNo qsemn) name = do
  oldName <- C.newEmptyMVar
  debugStop <- sendDebugMsg (QSemNLabel qsemnNo name oldName)
  C.takeMVar debugStop
  C.takeMVar oldName


labelSampleVar :: SampleVar a -> String -> IO String
labelSampleVar (SampleVar samplevarNo samplevar) name = do
  oldName <- C.newEmptyMVar
  debugStop <- sendDebugMsg (SampleVarLabel samplevarNo name oldName)
  C.takeMVar debugStop
  C.takeMVar oldName


labelThread :: String -> IO String
labelThread name = do
  oldName <- C.newEmptyMVar
  debugStop <- sendDebugMsg (ThreadLabel name oldName)
  C.takeMVar debugStop
  C.takeMVar oldName


forkIOLabel :: String -> IO () -> IO C.ThreadId
forkIOLabel name thread = 
  forkIO (do
    labelThread name
    thread
    )


forkIOLabelLine :: CodePosition -> String -> IO () -> IO C.ThreadId
forkIOLabelLine pos name thread = 
  forkIOLine pos (do
    labelThread name
    thread
    )