-------------------------------------------------------------------------------
-- Concurrent Haskell Stepper
--   BaseTypes Module
--
-------------------------------------------------------------------------------

module CHS.BaseTypes
  (
    IO(..),
    ActionType(..),
    Thread,
    Object(..),
    ThreadNo(..),
    ChanNo(..),
    MVarNo(..),
    QSemNo(..),
    QSemNNo(..),
    SampleVarNo(..),
    CodePosition(..),
    Result(..),
    Dependency(..)
  )
  where


-------------------------------------------------------------------------------
-- IMPORTS
-------------------------------------------------------------------------------

import Prelude hiding ( IO, print, putStr, getChar, getLine )
import qualified Prelude as P
import qualified Control.Concurrent as C

import CHD.BaseTypes


-------------------------------------------------------------------------------
-- TYPES
-------------------------------------------------------------------------------

data IO a = forall b. Bind (IO b) (b -> IO a)
	  | forall b. ConcAction Object (P.IO a) (P.IO (Maybe (a,b))) 
				 (b -> P.IO ())
	  | ForkAction (IO ()) (P.IO (a,C.ThreadId)) (P.IO (a,C.ThreadId))
	  | KillAction C.ThreadId (P.IO a) (P.IO a)
	  | SeqAction (ActionType a) (P.IO a)
	  | Return a

instance Monad IO where
  (>>=) = Bind
  return = Return

instance Show (IO a) where
  show (Bind iob ioa) = " (Bind " ++ show iob 
  show (ConcAction _ _ _ _) = "CA"
  show (ForkAction _ _ _) = "FA"
  show (KillAction _ _ _) = "KA"
  show (SeqAction _ _) = "SA"


data ActionType a = NonBackTrackable
		  | Executable
		  | Ignorable (P.IO a)
		  | MyThreadId (C.ThreadId -> P.IO a)


type Thread = (C.ThreadId,IO ())


data Result a = Stepped a
	      | Suspended
	      | Stop
	      | Fork Thread a
	      | Kill C.ThreadId a
	      | Terminate
	      | NotStepped a

data Dependency = Thread C.ThreadId
                | Object Object
                | NoDependency
  deriving (Eq,Show)
