-------------------------------------------------------------------------------
-- Concurrent Haskell Stepper
--   StepperIO Module
--   
-------------------------------------------------------------------------------

module CHS.StepperIO
  (
  SI.stdin,			-- :: Handle
  SI.stdout,			-- :: Handle
  SI.stderr,			-- :: Handle
  openFile,			-- :: FilePath -> IOMode -> IO Handle
  hClose,			-- :: Handle -> IO ()
  hFileSize,			-- :: Handle -> IO Integer
  hIsEOF,			-- :: Handle -> IO Bool
  isEOF,			-- :: IO Bool
  hSetBuffering,		-- :: Handle -> BufferMode -> IO ()
  hGetBuffering,		-- :: Handle -> IO BufferMode
  hFlush,			-- :: Handle -> IO ()
  hGetPosn,			-- :: Handle -> IO HandlePosn
  hSetPosn,			-- :: HandlePosn -> IO ()
  hSeek,			-- :: Handle -> SeekMode -> Integer -> IO ()
  hTell,			-- :: Handle -> IO Integer
  hWaitForInput,		-- :: Handle -> Int -> IO Bool
  hReady,			-- :: Handle -> IO Bool
  hGetChar,			-- :: Handle -> IO Char
  hGetLine,			-- :: Handle -> IO String
  hLookAhead,			-- :: Handle -> IO Char
  hGetContents,			-- :: Handle -> IO String
  hPutChar,			-- :: Handle -> Char -> IO ()
  hPutStr,			-- :: Handle -> String -> IO ()
  hPutStrLn,			-- :: Handle -> String -> IO ()
  hPrint,			-- :: (Show a) => Handle -> a -> IO ()
  hIsOpen,			-- :: Handle -> IO Bool
  hIsClosed,			-- :: Handle -> IO Bool
  hIsReadable,			-- :: Handle -> IO Bool
  hIsWritable,			-- :: Handle -> IO Bool
  hIsSeekable,			-- :: Handle -> IO Bool
  SI.isAlreadyExistsError,	-- :: IOError -> Bool
  SI.isDoesNotExistError,	-- :: IOError -> Bool
  SI.isAlreadyInUseError,	-- :: IOError -> Bool
  SI.isFullError,		-- :: IOError -> Bool
  SI.isEOFError,		-- :: IOError -> Bool
  SI.isIllegalOperation,	-- :: IOError -> Bool
  SI.isPermissionError,		-- :: IOError -> Bool
  SI.isUserError,		-- :: IOError -> Bool
  SI.ioeGetErrorString,		-- :: IOError -> String
  SI.ioeGetHandle,		-- :: IOError -> Maybe Handle
  SI.ioeGetFileName,		-- :: IOError -> Maybe FilePath
  try,				-- :: IO a -> IO (Either IOError a)
  IO,
  FilePath,
  IOError,
  ioError,			-- :: IOError -> IO a
  userError,			-- :: String -> IOError
  SI.catch,			-- :: IO a -> (IOError -> IO a) -> IO a
  interact,			-- :: (String -> String) -> IO ()
  putChar,			-- :: Char -> IO ()
  putStr,			-- :: String -> IO ()
  putStrLn,			-- :: String -> IO ()
  print,			-- :: (Show a) => a -> IO ()
  getChar,			-- :: IO Char
  getLine,			-- :: IO String
  getContents,			-- :: IO String
  readFile,			-- :: FilePath -> IO String
  writeFile,			-- :: FilePath -> String -> IO ()
  appendFile,			-- :: FilePath -> String -> IO ()
  readIO,			-- :: (Read a) => String -> IO a
  readLn,			-- :: (Read a) => IO a
  hPutBuf,			-- :: Handle -> Ptr a -> Int -> IO ()
  hGetBuf,			-- :: Handle -> Ptr a -> Int -> IO Int
  SI.fixIO,			-- :: (a -> IO a) -> IO a    -- !!
  hSetEcho,			-- :: Handle -> Bool -> IO ()
  hGetEcho,			-- :: Handle -> IO Bool
  hIsTerminalDevice,		-- :: Handle -> IO Bool,
  randomIO,			-- :: IO a
  getClockTime,			-- :: IO ST.ClockTime
  module System.Time,
  start,
  module Tcl,
  labelProgram,
  labelThread
  )
  where


-------------------------------------------------------------------------------
-- IMPORTS
-------------------------------------------------------------------------------

import Prelude hiding ( 
		       IO, 
		       putStr, 
		       readLn, 
		       readIO, 
		       appendFile, 
		       writeFile, 
		       readFile,
		       getContents,
		       getLine,
		       getChar,
		       print,
		       putStrLn,
		       putChar,
		       interact,
		       ioError,
		       catch
		      )
import qualified Prelude as P
import qualified ConcurrentDebug as CD
import qualified System.IO as SI
import qualified System.Random as SR
import qualified Control.Exception as CE
import qualified System.Time as ST
import System.Time hiding( getClockTime )
import qualified Tcl as TCL
import Tcl hiding( start )
import GHC.Exception ( catchException )
import Ptr

import CHS.BaseTypes


-------------------------------------------------------------------------------
-- TYPES
-------------------------------------------------------------------------------


-------------------------------------------------------------------------------
-- FUNCTIONS
-------------------------------------------------------------------------------

openFile :: FilePath -> SI.IOMode -> IO SI.Handle
openFile f i = SeqAction NonBackTrackable (SI.openFile f i)


hClose :: SI.Handle -> IO ()
hClose h = SeqAction (Ignorable (return ())) (SI.hClose h)


hFileSize :: SI.Handle -> IO Integer
hFileSize h = SeqAction Executable (SI.hFileSize h)


hIsEOF :: SI.Handle -> IO Bool
hIsEOF h = SeqAction Executable (SI.hIsEOF h)


isEOF :: IO Bool
isEOF = SeqAction Executable SI.isEOF


hSetBuffering :: SI.Handle -> SI.BufferMode -> IO ()
hSetBuffering h b = SeqAction (Ignorable (return ())) (SI.hSetBuffering h b)


hGetBuffering :: SI.Handle -> IO SI.BufferMode
hGetBuffering h = SeqAction NonBackTrackable (SI.hGetBuffering h)


hFlush :: SI.Handle -> IO ()
hFlush h = SeqAction (Ignorable (return ())) (SI.hFlush h)


hGetPosn :: SI.Handle -> IO SI.HandlePosn
hGetPosn h = SeqAction NonBackTrackable (SI.hGetPosn h)


hSetPosn :: SI.HandlePosn -> IO ()
hSetPosn h = SeqAction (Ignorable (return ())) (SI.hSetPosn h)


hSeek :: SI.Handle -> SI.SeekMode -> Integer -> IO ()
hSeek h s i = SeqAction Executable (SI.hSeek h s i)


hTell :: SI.Handle -> IO Integer
hTell h = SeqAction Executable (SI.hTell h)


hWaitForInput :: SI.Handle -> Int -> IO Bool
hWaitForInput h i = SeqAction NonBackTrackable (SI.hWaitForInput h i)


hReady :: SI.Handle -> IO Bool
hReady h = SeqAction Executable (SI.hReady h)


hGetChar :: SI.Handle -> IO Char
hGetChar h = SeqAction NonBackTrackable (SI.hGetChar h)


hGetLine :: SI.Handle -> IO String
hGetLine h = SeqAction NonBackTrackable (SI.hGetLine h)


hLookAhead :: SI.Handle -> IO Char
hLookAhead h = SeqAction NonBackTrackable (SI.hLookAhead h)


hGetContents :: SI.Handle -> IO String
hGetContents h = SeqAction NonBackTrackable (SI.hGetContents h)


hPutChar :: SI.Handle -> Char -> IO ()
hPutChar h c = SeqAction (Ignorable (return ())) (SI.hPutChar h c)


hPutStr :: SI.Handle -> String -> IO ()
hPutStr handle str = SeqAction (Ignorable (return ())) (SI.hPutStr handle str)


hPutStrLn :: SI.Handle -> String -> IO ()
hPutStrLn h str = SeqAction (Ignorable (return ())) (SI.hPutStrLn h str)


hPrint :: (Show a) => SI.Handle -> a -> IO ()
hPrint h = hPutStr h . show


hIsOpen :: SI.Handle -> IO Bool
hIsOpen h = SeqAction Executable (SI.hIsOpen h)


hIsClosed :: SI.Handle -> IO Bool
hIsClosed h = SeqAction Executable (SI.hIsClosed h)


hIsReadable :: SI.Handle -> IO Bool
hIsReadable h = SeqAction Executable (SI.hIsReadable h)


hIsWritable :: SI.Handle -> IO Bool
hIsWritable h = SeqAction Executable (SI.hIsWritable h)


hIsSeekable :: SI.Handle -> IO Bool
hIsSeekable h = SeqAction Executable (SI.hIsSeekable h)


try :: IO a -> IO (Either IOError a)
try io = Bind io (\b -> SeqAction Executable 
				  (do ei <- SI.try (return b)
				      return $ either (\x->Left x) 
						      (\x->Right x) ei))


ioError :: IOError -> IO a
ioError ie =  SeqAction Executable (SI.ioError ie)


catch :: IO a -> (IOError -> IO a) -> IO a
catch io fio = Bind (try io) 
		    (\either -> case either of
				     Left ioe -> fio ioe
				     Right x  -> Return x)


interact :: (String -> String) -> IO ()
interact f = SeqAction (Ignorable (return ())) (SI.interact f)


putChar :: Char -> IO ()
putChar char = SeqAction (Ignorable (return ())) (P.putChar char)


putStr :: String -> IO ()
putStr str = SeqAction (Ignorable (return ())) (P.putStr str)


putStrLn :: String -> IO ()
putStrLn str = SeqAction (Ignorable (return ())) (P.putStrLn str)


print :: (Show a) => a -> IO ()
print = putStr . show


getChar :: IO Char
getChar = SeqAction NonBackTrackable SI.getChar


getLine :: IO String
getLine = SeqAction NonBackTrackable SI.getLine


getContents :: IO String
getContents = SeqAction NonBackTrackable SI.getContents


readFile :: FilePath -> IO String
readFile f = SeqAction NonBackTrackable (SI.readFile f)


writeFile :: FilePath -> String -> IO ()
writeFile f s = SeqAction NonBackTrackable (SI.writeFile f s) 


appendFile :: FilePath -> String -> IO ()
appendFile f s = SeqAction NonBackTrackable (SI.appendFile f s)


readIO :: (Read a) => String -> IO a
readIO s = SeqAction NonBackTrackable (SI.readIO s)


readLn :: (Read a) => IO a
readLn = SeqAction NonBackTrackable SI.readLn


hPutBuf :: SI.Handle -> Ptr a -> Int -> IO ()
hPutBuf h p i = SeqAction (Ignorable (return ())) (SI.hPutBuf h p i)


hGetBuf :: SI.Handle -> Ptr a -> Int -> IO Int
hGetBuf h p i = SeqAction NonBackTrackable (SI.hGetBuf h p i)


--fixIO :: (a -> IO a) -> IO a
--fixIO f = SeqAction Executable (SI.fixIO (executeIO . f))


hSetEcho :: SI.Handle -> Bool -> IO ()
hSetEcho h b =  SeqAction (Ignorable (return ())) (SI.hSetEcho h b)


hGetEcho :: SI.Handle -> IO Bool
hGetEcho h = SeqAction NonBackTrackable (SI.hGetEcho h)


hIsTerminalDevice :: SI.Handle -> IO Bool
hIsTerminalDevice h = SeqAction Executable (SI.hIsTerminalDevice h)


randomIO :: SR.Random a => IO a
randomIO = SeqAction NonBackTrackable (SR.randomIO)


-- System.Time

getClockTime :: IO ClockTime
getClockTime = SeqAction Executable ST.getClockTime


-- TclHaskell

start :: GUI () -> IO ()
start g = SeqAction Executable (TCL.start g)


-- proc :: IO () -> GUI ()
-- proc io = 




labelProgram :: String -> IO () 
labelProgram label = SeqAction (Ignorable (return ())) (CD.labelProgram label)


labelThread :: String -> IO String
labelThread label =
  SeqAction (Ignorable (return "")) (CD.labelThread label)