-------------------------------------------------------------------------------
-- Concurrent Haskell Stepper
--   StepperMVar Module
--   
-------------------------------------------------------------------------------

module CHS.StepperMVar
  (
   MVar(..),
   newEmptyMVar,	 -- :: IO (MVar a)
   newEmptyMVarLine,	 -- :: CodePosition -> IO (MVar a)
   newMVar,		 -- :: a -> IO (MVar a)
   newMVarLine,		 -- :: CodePosition -> a -> IO (MVar a)
   newMVarLabel,	 -- :: String -> a -> IO (MVar a)
   newMVarLabelLine,	 -- :: CodePosition -> String -> a -> IO (MVar a)
   takeMVar,		 -- :: MVar a -> IO a
   takeMVarLine,	 -- :: CodePosition -> MVar a -> IO a
   putMVar,		 -- :: MVar a -> a -> IO ()
   putMVarLine,		 -- :: CodePosition -> MVar a -> a -> IO () 
   putMVarLabel,	 -- :: String -> MVar a -> a -> IO ()
   putMVarLabelLine,	 -- :: CodePosition -> String -> MVar a -> a -> IO ()
   readMVar,		 -- :: MVar a -> IO a
   readMVarLine,	 -- :: CodePosition -> MVar a -> IO a 
   swapMVar,		 -- :: MVar a -> a -> IO a
   swapMVarLine,	 -- :: CodePosition -> MVar a -> a -> IO a
   swapMVarLabel,	 -- :: MVar a -> a -> String -> IO a
   swapMVarLabelLine,	 -- :: CodePosition -> String -> MVar a -> a -> IO a
   tryTakeMVar,		 -- :: MVar a -> IO (Maybe a)
   tryTakeMVarLine,	 -- :: CodePosition -> MVar a -> IO (Maybe a)
   tryPutMVar,		 -- :: MVar a -> a -> IO Bool
   tryPutMVarLine,	 -- :: CodePosition -> MVar a -> a -> IO Bool
   tryPutMVarLabel,	 -- :: String -> MVar a -> a -> IO Bool
   tryPutMVarLabelLine,	 -- :: CodePosition -> String -> MVar a -> a -> IO Bool
   isEmptyMVar,		 -- :: MVar a -> IO Bool
   labelMVar		 -- :: MVar a -> String -> IO String
  )
  where


-------------------------------------------------------------------------------
-- IMPORTS
-------------------------------------------------------------------------------

import Prelude hiding ( IO )
import qualified Prelude as P
import qualified Control.Concurrent as C
import qualified ConcurrentDebug as CD

import CHS.BaseTypes
import CHS.ComMVars

import System.IO.Unsafe


-------------------------------------------------------------------------------
-- TYPES
-------------------------------------------------------------------------------

data MVar a = MVar (CD.MVar a) (C.MVar a) (C.MVar ()) (C.MVar ())


------------------------------------------------------------------------------
-- FUNCTIONS
-------------------------------------------------------------------------------

newMVarNoMVar :: C.MVar MVarNo
newMVarNoMVar = unsafePerformIO $ C.newMVar (MVarNo (-1))


newEmptyMVar :: IO (MVar a)
newEmptyMVar = newEmptyMVarLine NoPosition


newEmptyMVarLine :: CodePosition -> IO (MVar a)
newEmptyMVarLine pos =
  ConcAction (MVarObj (MVarNo (-1)))
	     (do cdMVar <- CD.newEmptyMVarLine pos
	         C.putMVar breakMVar ()
		 C.takeMVar finishedBreakMVar
		 searchMVar <- C.newEmptyMVar
		 putStop <- C.newEmptyMVar
		 takeStop <- C.newEmptyMVar
		 return $ MVar cdMVar searchMVar putStop takeStop)
	     (do cMVar <- C.newEmptyMVar
		 searchMVar <- C.newEmptyMVar
		 putStop <- C.newEmptyMVar
		 takeStop <- C.newEmptyMVar
		 mvarNo <- C.takeMVar newMVarNoMVar
		 let cdMVar = CD.MVar mvarNo cMVar
		     mvar = MVar cdMVar searchMVar putStop takeStop
		 C.putMVar newMVarNoMVar (pred mvarNo)
		 return $ Just (mvar,mvar))
             (\_ -> do mvarNo <- C.takeMVar newMVarNoMVar
		       C.putMVar newMVarNoMVar (succ mvarNo))


newMVar :: a -> IO (MVar a)
newMVar = newMVarLabelLine NoPosition ""


newMVarLine :: CodePosition -> a -> IO (MVar a)
newMVarLine pos = newMVarLabelLine pos ""


newMVarLabel :: String -> a -> IO (MVar a)
newMVarLabel = newMVarLabelLine NoPosition


newMVarLabelLine :: CodePosition -> String -> a -> IO (MVar a)
newMVarLabelLine pos label x = 
  ConcAction (MVarObj (MVarNo (-1)))
	     (do cdMVar <- CD.newMVarLabelLine pos label x
	         C.putMVar breakMVar ()
		 C.takeMVar finishedBreakMVar
		 searchMVar <- C.newMVar x
		 putStop <- C.newMVar ()
		 takeStop <- C.newMVar ()
		 return $ MVar cdMVar searchMVar putStop takeStop)
	     (do cMVar <- C.newMVar x
		 searchMVar <- C.newMVar x
		 putStop <- C.newMVar ()
		 takeStop <- C.newMVar ()
		 mvarNo <- C.takeMVar newMVarNoMVar
		 let cdMVar = CD.MVar mvarNo cMVar
		     mvar = MVar cdMVar searchMVar putStop takeStop
		 C.putMVar newMVarNoMVar (pred mvarNo)
		 return $ Just (mvar,mvar))
             (\_ -> do mvarNo <- C.takeMVar newMVarNoMVar
		       C.putMVar newMVarNoMVar (succ mvarNo))


putMVar :: MVar a -> a -> IO ()
putMVar = putMVarLine NoPosition


putMVarLine :: CodePosition -> MVar a -> a -> IO ()
putMVarLine pos = putMVarLabelLine pos ""


putMVarLabel :: String -> MVar a -> a -> IO ()
putMVarLabel = putMVarLabelLine NoPosition


putMVarLabelLine :: CodePosition -> String -> MVar a -> a -> IO ()
putMVarLabelLine pos label 
	(MVar cdMVar@(CD.MVar mvarNo _) searchMVar putStop takeStop) x=
  ConcAction (MVarObj mvarNo) 
	     (do CD.putMVarLabelLine pos label cdMVar x
		 C.putMVar putStop ()
		 C.putMVar breakMVar ()
		 C.putMVar takeStop ()
		 C.takeMVar finishedBreakMVar
		 C.putMVar searchMVar x)
	     (do b <- C.isEmptyMVar searchMVar
		 if b
		    then do C.putMVar searchMVar x
			    return $ Just ((),())
		    else return Nothing)
             (\_ -> do C.takeMVar searchMVar
		       return ())


takeMVar :: MVar a -> IO a
takeMVar = takeMVarLine NoPosition


takeMVarLine :: CodePosition -> MVar a -> IO a
takeMVarLine pos (MVar cdMVar@(CD.MVar mvarNo _) searchMVar putStop takeStop) =
  ConcAction (MVarObj mvarNo)
	     (do CD.takeMVarLine pos cdMVar
		 C.takeMVar takeStop
		 C.putMVar breakMVar ()
		 C.takeMVar putStop
		 C.takeMVar finishedBreakMVar
		 C.takeMVar searchMVar)
	     (do b <- C.isEmptyMVar searchMVar
                 if b
                    then return Nothing
                    else do v <- C.takeMVar searchMVar
			    return $ Just (v,v))
             (\value -> C.putMVar searchMVar value)


readMVar :: MVar a -> IO a
readMVar = readMVarLine NoPosition


readMVarLine :: CodePosition -> MVar a -> IO a
readMVarLine pos (MVar cdMVar@(CD.MVar mvarNo _) searchMVar _ takeStop) = 
  ConcAction (MVarObj mvarNo)
	     (do CD.readMVarLine pos cdMVar
		 C.takeMVar takeStop
		 C.putMVar breakMVar ()
		 C.putMVar takeStop ()
		 C.takeMVar finishedBreakMVar
		 C.readMVar searchMVar)
             (do b <- C.isEmptyMVar searchMVar
		 if b
                    then return Nothing
                    else do v <- C.readMVar searchMVar
			    return $ Just (v,v))
             (\_ -> return ())


swapMVar :: MVar a -> a -> IO a
swapMVar = swapMVarLabelLine NoPosition ""


swapMVarLine :: CodePosition -> MVar a -> a -> IO a
swapMVarLine pos = swapMVarLabelLine pos ""


swapMVarLabel :: String -> MVar a -> a -> IO a
swapMVarLabel = swapMVarLabelLine NoPosition


swapMVarLabelLine :: CodePosition -> String -> MVar a -> a -> IO a
swapMVarLabelLine pos label (MVar cdMVar@(CD.MVar mvarNo _) searchMVar _ _) x =
  ConcAction (MVarObj mvarNo)
	     (do CD.swapMVarLabelLine pos label cdMVar x
		 C.putMVar breakMVar ()
		 C.takeMVar finishedBreakMVar
		 C.swapMVar searchMVar x)
	     (do b <- C.isEmptyMVar searchMVar
		 if b
		    then return Nothing
		    else do v <- C.swapMVar searchMVar x
			    return $ Just (v,v))
             (\v -> do C.swapMVar searchMVar v
		       return ())


tryTakeMVar :: MVar a -> IO (Maybe a)
tryTakeMVar = tryTakeMVarLine NoPosition


tryTakeMVarLine :: CodePosition -> MVar a -> IO (Maybe a)
tryTakeMVarLine pos (MVar cdMVar@(CD.MVar mvarNo _) searchMVar putStop 
		takeStop) =
  ConcAction (MVarObj mvarNo)
	     (do m <- CD.tryTakeMVarLine pos cdMVar
		 maybe (return ()) (\_ -> C.takeMVar takeStop) m
		 C.putMVar breakMVar ()
		 maybe (return ()) (\_ -> C.takeMVar putStop) m
		 C.takeMVar finishedBreakMVar
		 C.tryTakeMVar searchMVar)
	     (do v <- C.tryTakeMVar searchMVar
		 return $ Just (v,v))
             (maybe (return ()) (C.putMVar searchMVar))


tryPutMVar :: MVar a -> a -> IO Bool
tryPutMVar = tryPutMVarLabelLine NoPosition ""


tryPutMVarLine :: CodePosition -> MVar a -> a -> IO Bool
tryPutMVarLine pos = tryPutMVarLabelLine pos ""


tryPutMVarLabel :: String -> MVar a -> a -> IO Bool
tryPutMVarLabel = tryPutMVarLabelLine NoPosition


tryPutMVarLabelLine :: CodePosition -> String -> MVar a -> a -> IO Bool
tryPutMVarLabelLine pos label 
	(MVar cdMVar@(CD.MVar mvarNo _) searchMVar putStop takeStop) x =
  ConcAction (MVarObj mvarNo) 
	     (do b <- CD.tryPutMVarLabelLine pos label cdMVar x
		 if b
		    then C.putMVar putStop ()
		    else return ()
		 C.putMVar breakMVar ()
		 if b
		    then C.putMVar takeStop ()
		    else return ()
		 C.takeMVar finishedBreakMVar
		 C.tryPutMVar searchMVar x)
	     (do b <- C.isEmptyMVar searchMVar
		 if b
		    then do v <- C.tryPutMVar searchMVar x
			    return $ Just (v,v)
		    else return Nothing)
             (\b -> if b
		       then do C.tryTakeMVar searchMVar
			       return ()
		       else return ())


isEmptyMVar :: MVar a -> IO Bool
isEmptyMVar (MVar cdMVar@(CD.MVar mvarNo _) searchMVar _ _) = 
  ConcAction (MVarObj mvarNo)
	     (do CD.isEmptyMVar cdMVar
		 C.putMVar breakMVar ()
		 C.takeMVar finishedBreakMVar
		 C.isEmptyMVar searchMVar)
	     (do b <- C.isEmptyMVar searchMVar
		 return $ Just (b,b))
             (\_ -> return ())


labelMVar :: MVar a -> String -> IO String
labelMVar (MVar cdMVar _ _ _) label = 
  SeqAction (Ignorable (return label)) (CD.labelMVar cdMVar label)