-------------------------------------------------------------------------------
-- Concurrent Haskell Stepper
--   StepperQSem Module
--   
-------------------------------------------------------------------------------

module CHS.StepperQSem
  (
   QSem,
   newQSem,		-- :: Int -> IO QSem
   newQSemLine,		-- :: CodePosition -> Int -> IO QSem
   signalQSem,		-- :: QSem -> IO ()
   signalQSemLine,	-- :: CodePosition -> QSem -> IO ()
   waitQSem,		-- :: QSem -> IO ()
   waitQSemLine		-- :: CodePosition -> QSem -> IO ()   
  )
  where


-------------------------------------------------------------------------------
-- IMPORTS
-------------------------------------------------------------------------------

import Prelude hiding ( IO, print, putStr, putStrLn, getChar, getLine )
import qualified Prelude as P
import qualified Control.Concurrent as C
import qualified ConcurrentDebug as CD
import qualified System.IO as SI

import CHS.BaseTypes
import CHS.ComMVars

import System.IO.Unsafe

import qualified QSem as Q


-------------------------------------------------------------------------------
-- TYPES
-------------------------------------------------------------------------------

data QSem = QSem CD.QSem Q.QSem Q.QSem


-------------------------------------------------------------------------------
-- FUNCTIONS
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
-- QSem-Aktionen
-------------------------------------------------------------------------------

newQSemNoMVar :: C.MVar QSemNo
newQSemNoMVar = unsafePerformIO $ C.newMVar (QSemNo (-1))


newQSem :: Int -> IO QSem
newQSem x = newQSemLine NoPosition x


newQSemLine :: CodePosition -> Int -> IO QSem
newQSemLine pos x = 
  ConcAction (QSemObj (QSemNo (-1))) 
	     (do cdQSem <- CD.newQSemLine pos x
	         C.putMVar breakMVar ()
		 C.takeMVar finishedBreakMVar
		 searchQSem <- Q.newQSem x
		 waitStop <- Q.newQSem x
		 return $ QSem cdQSem searchQSem waitStop)
	     (do qsem <- C.newQSem x
		 searchQSem2 <- Q.newQSem x
		 waitStop2 <- Q.newQSem x
		 qsemNo <- C.takeMVar newQSemNoMVar
		 let cdQSem2 = CD.QSem qsemNo qsem
		     qsem2 = QSem cdQSem2 searchQSem2 waitStop2
		 C.putMVar newQSemNoMVar (pred qsemNo)
		 return $ Just (qsem2,qsem2))
             (\_ -> do qsemNo <- C.takeMVar newQSemNoMVar
		       C.putMVar newQSemNoMVar (succ qsemNo))


signalQSem :: QSem -> IO ()
signalQSem = signalQSemLine NoPosition


signalQSemLine :: CodePosition -> QSem -> IO ()
signalQSemLine pos (QSem cdQSem@(CD.QSem qsemNo _) searchQSem waitStop) =
  ConcAction (QSemObj qsemNo) 
	     (do CD.signalQSemLine pos cdQSem
		 C.putMVar breakMVar ()
		 Q.signalQSem waitStop
		 C.takeMVar finishedBreakMVar
		 Q.signalQSem searchQSem)
	     (do Q.signalQSem searchQSem
		 return $ Just ((),()))
	     (\_ -> Q.waitQSem searchQSem)


waitQSem :: QSem -> IO ()
waitQSem = waitQSemLine NoPosition


waitQSemLine :: CodePosition -> QSem -> IO ()
waitQSemLine pos (QSem cdQSem@(CD.QSem qsemNo _) searchQSem waitStop) =
  ConcAction (QSemObj qsemNo)
	     (do CD.waitQSemLine pos cdQSem
		 Q.waitQSem waitStop
		 C.putMVar breakMVar ()
		 C.takeMVar finishedBreakMVar
		 Q.waitQSem searchQSem)
	     (do q <- Q.getQuantityQSem searchQSem
                 if q>0
                    then do Q.waitQSem searchQSem
			    return $ Just ((),())
		    else return Nothing)
             (\_ -> Q.signalQSem searchQSem)